@@ -56,7 +56,8 @@ def date_filter_parser(query: Query, field: Column, filter_string: str) -> Query
         value = date(2000 + int(value[4:6]), int(value[2:4]), int(value[0:2]))
     elif len(value) == 8:
         value = date(int(value[4:8]), int(value[2:4]), int(value[0:2]))
-    # else means isodate
+    else:  # ISO date
+        value = date(*(int(v) for v in value.split('-')))
     return math_query(query, field, tokens[0], value)
 
 
